from django.conf import settings
#from django.http import QueryDict
from django.core.exceptions import MiddlewareNotUsed
import base64
import hashlib
import hmac
import re
import os

from marketplace import logger


class MockMiddleware(object):
    """
Use this middleware to mock exactly what the HubSpot Marketplace would do with
your requests/responses locally.  This allows you to test your app locally and
hit it just as it would be hit in production.  To use, you'll need to:

1) Make sure you've already got your AuthMiddlware all setup -- see docs for
that guy if you don't know how.

2) Place this class in your middlewares list in your settings.py BEFORE the
AuthMiddleware.

3) Give the middleware enough information to correctly mock out how things
would go down in production, adding something like this to your settings.py:

    HUBSPOT_MARKETPLACE_MOCK = {
        'slug': 'yourappslug',
        'app': {
            'name': 'YourAppName', # override with your app name
            'callback_url': 'http://localhost:8000',
        },
        'user': {
            'id': 9999999,  # fake user_id
        }
    }

This middleware assumes that anything coming in on /market/HUBID/canvas/SLUG
should seem as a marketplace request should seem, and it appropriately adds all
the attributes to the request that would be added in production.

It also rewrites all your hsml on the response as you would expect them
rewritten in production.

NOTE-- this mock will always have to play catchup with the marketplace.  As they
add new features to the marketplace, they need to also be mocked here.  Please
feel free to contribute to this effort.
    """

    MOCK_SETTINGS_DEFAULTS = {
        'slug': 'yourslug', # override with your actual slug
        'app': {
            'name': 'MyAppName', # override with your app name
            'callback_url': 'http://localhost:8000',
        },
        'user': {
            'id': 9999999,  # fake user_id so we have something
        },
        'caller': 'Hubspot Marketplace',  # don't expect this overridden
    }


    def __init__(self):
        super(MockMiddleware,self).__init__()
        mock = getattr(settings, 'HUBSPOT_MARKETPLACE_MOCK', {})
        mock_safety = getattr(settings, 'HUBSPOT_MARKETPLACE_MOCK_SAFETY', None)
        auth = getattr(settings, 'HUBSPOT_MARKETPLACE_AUTH', {})
        secret = auth.get('secret_key')
        self.log = logger.get_log(__name__)
        if not mock or mock and not secret or mock_safety and not os.environ.get(mock_safety):
            self.log.info(
                    'MockMiddleware has been turned off for all requests')
            raise MiddlewareNotUsed
        payload = 'payload'
        digest = hmac.new(secret, payload, hashlib.sha1).digest()
        self.signature = '.'.join(
                [self.base64_url_encode_for_real(s) 
                    for s in [digest, payload]])

        self.slug = mock.get('slug')
        if not self.slug:
            raise KeyError("Missing slug definition in MockMiddleware")

        self.prefix_path_re = re.compile('/market/(\d+)/canvas/%s'%self.slug)
        self.body_re = re.compile(r'<body>(.*?)</body>',re.DOTALL)
        self.head_re = re.compile(r'<hs:head>(.*?)</hs:head>',re.DOTALL)
        self.link_re = re.compile(r'<hs:link (.*?)/?>')
        self.script_re = re.compile(r'<hs:script(.*?)>(.*?)</hs:script>', re.DOTALL)
        self.title_re = re.compile(r'<hs:title(.*?)>(.*?)</hs:title>')
        self.form_re = re.compile(r'(<form\s.*?)action="(/.*?)"')

# keeping this kicking around cuz we'll likely uncomment when marketplace fixes
# how this works
#        self.anchor_re = re.compile(r'(<a\s.*?)href="(/.*?)"')

        path = os.path.join(os.path.dirname(__file__),'mock.html')
        self.wrapper = open(path).read()

        self.build_static_params(mock)

        self.log.info('HubSpot Marketplace Mock Canvas Middleware Activated')


    def build_static_params(self, mock):
        """
        Builds the params in a very forgiving way.  Necessary for backward
        compatibility
        """

        mapping = {
            'caller': 'caller',
            'user_id': 'user.id',
            'user.email': 'user.email',
            'user.first_name': 'user.first_name',
            'user.last_name': 'user.last_name',
            'app.name': 'app.name',
            'app.callbackUrl': 'app.callback_url',
            'app.pageUrl': 'page.url',
        }
        base = {}
        defaults = self.__class__.MOCK_SETTINGS_DEFAULTS
        for k,v in mapping.iteritems():
            val = None
            for store in [mock, defaults]:
                if not val:
                    val = store.get(k) or store.get(v)
                    for l in [k,v]:
                        if not val and '.' in l:
                            parts = l.split('.')
                            temp = store.get(parts[0],{})
                            val = val or temp.get(parts[1])

            base['hubspot.marketplace.%s'%k] = str(val or '')
        base['hubspot.marketplace.signature'] = self.signature
        base['hubspot.marketplace.is_mock'] = 'true'
        self.base = base


    def process_request(self, request):
        url_match = self.prefix_path_re.match(request.path)
        if not url_match: # this isn't a url that needs marketplace mocking
            return 

        # make the relevant dictionary mutable
        params = getattr(request, request.method).copy()

        hub_id = int(url_match.group(1))

        for k,v in self.base.iteritems():
            params.appendlist(k,v)
        params.appendlist('hubspot.marketplace.portal_id', str(hub_id))
        params.appendlist('hubspot.marketplace.app.canvasUrl', 
                "http://%s/market/%s/canvas/%s/" % 
                (request.get_host(), hub_id, self.slug))
        params.appendlist('hubspot.marketplace.app.pageUrl', str(request.path))

        setattr(request, request.method, params)

        callback_url = self.base['hubspot.marketplace.app.callbackUrl']
        callback_path = (callback_url.split('//',1)+[''])[1] or callback_url
        callback_prefix = (callback_path.split('/',1)+[''])[1]
        callback_prefix = callback_prefix and '/%s'%callback_prefix

        request.path = self.prefix_path_re.sub(callback_prefix, request.path, 1)
        request.path_info = self.prefix_path_re.sub(callback_prefix, request.path_info, 1)


    def process_response(self, request, response):
        marketplace = request and getattr(request, 'marketplace', None)
        if marketplace:
            if self.body_re.search(response.content):
                innards = self.body_re.findall(response.content)[0]
                head = ''
                for link in self.link_re.findall(innards):
                    head += "\n<link %s />"%link
                for head_content in self.head_re.findall(innards):
                    head += "\n%s" % head_content
                bottom = ''
                for match_tuple in self.script_re.findall(innards):
                    bottom += "\n<script%s>%s</script>" % (match_tuple[0], match_tuple[1])
                innards = self.link_re.sub('',innards)
                innards = self.head_re.sub('',innards)
                innards = self.title_re.sub('',innards)
                innards = self.script_re.sub('',innards)
                innards = self.form_re.sub(r'\1action="/market/%s/canvas/%s\2"' %
                        (marketplace.hub_id,self.slug), innards)

    # keeping this kicking around cuz we'll likely uncomment when marketplace fixes
    # how this works
                    #innards = self.anchor_re.sub(r'\1href="/market/%s/canvas/%s\2"' %
                            #(marketplace.hub_id,self.slug), innards)

                content = self.wrapper
                content = content.replace('[[HEAD_CONTENTS]]',head)
                content = content.replace('[[BODY_CONTENTS]]',innards)
                content = content.replace('[[BOTTOM_BODY_CONTENTS]]',bottom)
                response.content = content
            #else:
##                print vars(response).keys()
                #head_style_re = re.compile(
                        #r'<style type="text/css">(.*?)</style>',
                        #re.DOTALL)
                #head_script_re = re.compile(
                        #r'<script type="text/javascript">(.*?)</script>',
                        #re.DOTALL)
                #if self.body_re.search(response.content):
                    #innards = self.body_re.findall(response.content)[0]
                    #for style in head_style_re.findall(response.content):
                        #innards = ('<style type="text/css">%s</style>'%style +
                        #innards)
                    #for script in head_script_re.findall(response.content):
                        #innards = ('<script type="text/javascript">%s</script>' % script
                                #+ innards)                          
                    #content = self.wrapper
                    #content = content.replace('[[HEAD_CONTENTS]]','')
                    #content = content.replace('[[BODY_CONTENTS]]','<div class="asdf">%s</div>'%innards)
                    #content = content.replace('[[BOTTOM_BODY_CONTENTS]]','')
                    #response.content = content

        return response




    def base64_url_encode_for_real(self, decoded_s):
        """
        base64 library decided to leave '=' chars still kicking around, and was
        still bold enough to call their method 'urlsafe' -- okaaaaay...  This
        method is urlsafe for real, and matches what the marketplace is
        expecting
        """
        return base64.urlsafe_b64encode(decoded_s).split('=',1)[0]








import re
import os
from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
from marketplace import logger


class ErrorGogglesMiddleware(object):
    """
Use this middleware to make django error stack traces show up properly inside
the marketplace wrapper.  If you've ever had to deal with a stack trace
without this middleware, you know it can get very hard to decipiher what the
issue is.

This middleware will only attempt to reformat for the marketplace when the
django app is in DEBUG mode.  Otherwise it won't attempt anything
    """


    def __init__(self):
        super(ErrorGogglesMiddleware,self).__init__()
        self.log = logger.get_log(__name__)
        debug_mode = getattr(settings, 'DEBUG', False)
        if not debug_mode:
            self.log.info('ErrorGogglesMiddleware has been turned off for all requests because we are not in DEBUG mode')
            raise MiddlewareNotUsed
        self.head_re = re.compile(r'<head>(.*?)</head>',re.DOTALL)
        self.body_re = re.compile(r'<body>(.*?)</body>',re.DOTALL)
        self.style_re = re.compile(r'<style type="text/css">(.*?)</style>', re.DOTALL)
        self.script_re = re.compile(r'<script type="text/javascript">(.*?)</script>', re.DOTALL)
        self.reset_styles = open(os.path.join(os.path.dirname(__file__),'error_goggles_reset.css')).read()
        self.log.info('ErrorGogglesMiddleware has been activated')

    def process_response(self, request, response):
        marketplace = request and getattr(request, 'marketplace', None)
        if marketplace and response.status_code >= 400:
            if self.head_re.search(response.content) and self.body_re.search(response.content):
                head = self.head_re.findall(response.content)[0]
                body = self.body_re.findall(response.content)[0]
                body = '<div class="hsmpdjerr">%s</div>' % body
                for style in self.style_re.findall(head):
                    body = '<style type="text/css">%s</style>'%style + body
                body = '<style type="text/css">%s</style>'%self.reset_styles + body
                for script in self.script_re.findall(head):
                    body = '<script type="text/javascript">%s</script>' % script + body
                response.content = self.body_re.sub('<body>%s</body>'%body, response.content)
        return response





from django.conf import settings
from django.http import HttpResponse
import logger

"""
View method decorators relevant to the hubspot marketplace
"""


AUTH_ERROR = """
This request did not have a proper HubSpot Marketplace authentication
signature!

This particular view was decorated so as to expect a valid HubSpot Marketplace
signature, however, the signature on this request was either missing,
malformed, or wrong.

Perhaps you incorrectly decorated this view as a marketplace view?  Or, perhaps
this request did not originate from HubSpot!?

If you're running locally, use the MockMiddleware to mock out all the
interactions you would be having with the marketplace in Production.  This will
give you the closest experience to riding the marketplace during your local
development.
""".strip()


def marketplace(function=None):
    """
    Use this decorator when you want to ensure marketplace authentication
    """

    def _dec(view_func):
        def _view(request, *args, **kwargs):
            authenticate = getattr(settings, 'HUBSPOT_MARKETPLACE_AUTH', {}).get('SECRET_KEY', False)
            if authenticate and not getattr(request, 'marketplace', None):
                logger.get_log('marketplace_decorator').error(AUTH_ERROR)
                return HttpResponse(status=401)
            else:
                return view_func(request, *args, **kwargs)
        _view.__name__ = view_func.__name__
        _view.__dict__ = view_func.__dict__
        _view.__doc__ = view_func.__doc__
        return _view
    
    if function is None:
        return _dec
    else:
        return _dec(function)




from request_supplement import RequestSupplement
from middleware.auth import AuthMiddleware
from middleware.mock import MockMiddleware
from middleware.anchor_fix import AnchorFixMiddleware
from middleware.error_goggles import ErrorGogglesMiddleware



import logging

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

def get_log(name):
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())
    return logger




import re
#from marketplace import logger


class AnchorFixMiddleware(object):
    """
Use this middleware to make absolute paths in anchor tags be prepended with a
protocol relative url.  This can be used as a fix for the current marketplace
wonkiness where it doesn't rewrite absolute urls in an anchor tag

look here for more info:
http://docs.hubapi.com/wiki/Link_Rewriting

This middleware essentially turns the box pointed to by <a> and by "Absolute
Url" into something useful.
    """

    def __init__(self):
        super(AnchorFixMiddleware,self).__init__()
        self.anchor_re = re.compile(r'(<a\s.*?)href="(/.*?)"')

    def process_response(self, request, response):
        marketplace = request and getattr(request, 'marketplace', None)
        if marketplace and getattr(marketplace,'base_url',None) and response.status_code==200:
            content = response.content.decode('utf-8')
            content = self.anchor_re.sub(r'\1href="%s\2"' % marketplace.base_url[0:-1], content)
            response.content = content.encode('utf-8')
        return response






from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
import hashlib
import hmac
import base64
from marketplace import logger
from marketplace import RequestSupplement


class AuthMiddleware(object):
    """
Use this middleware to ensure requests are coming from HubSpot and that they
are intended for your app.  To use, you'll need to:

1) Place this class in your middlewares list in you settings.py

2) Let the middleware know about your secret, adding something like this to
your settings.py:

    HUBSPOT_MARKETPLACE_AUTH = { 
        'secret_key': 'hubspot-issued-secret-key-here' 
    }

This middleware assumes that you may also want to allow some non-HubSpot
Marketplace requests to pass through to your web app as well.  So, although
authentication is established at this layer, it isn't enforced unless you've
applied the @marketplace decorator on your view function.  And the code doesn't
enforce it until it executes that decorator (which happens after all the
middlewares have executed).
    """

    DEACTIVATION_NOTICE = """
HubSpot marketplace request authentication has been deactivated for all
requests!  (because there was no secret key specified in settings.py.  Gotta
specify there if you wanna turn on authentication!)
""".strip()

    SIGNATURE_KEY = 'hubspot.marketplace.signature'
            
    def __init__(self):
        super(AuthMiddleware, self).__init__()
        self.log = logger.get_log(__name__)
        auth = getattr(settings, 'HUBSPOT_MARKETPLACE_AUTH', {})
        self.secret = auth.get('secret_key')
        if not self.secret:
            self.log.warn(self.__class__.DEACTIVATION_NOTICE)
            raise MiddlewareNotUsed

    def process_request(self, request):
        """adds MarketPlaceInfo object to request at request.marketplace"""
        signature = request.REQUEST.get(self.__class__.SIGNATURE_KEY, '').strip()
        if self.is_request_authentic(signature):
            request.marketplace = RequestSupplement(request)

    def is_request_authentic(self, signature):
        """ensures this request was issued by HubSpot"""
        signature = str(signature)  # convert from unicode
        digest, payload = [
                self.base64_url_decode_for_real(s) 
                    for s in (signature+'.').split('.')[0:2]]
        return (digest and payload and 
                digest == hmac.new(self.secret, payload, hashlib.sha1).digest())


    def base64_url_decode_for_real(self, encoded_s):
        """
        base64 library decided to leave '=' chars still kicking around, and was
        still bold enough to call their method 'urlsafe' -- okaaaaay...  This
        method is urlsafe for real, and matches what the marketplace is
        expecting 
        """
        return base64.urlsafe_b64decode(encoded_s + '=' * (4 - len(encoded_s) % 4))





import traceback
from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
from marketplace import logger


class DebugModeLoggingMiddleware(object):
    """
Use this middleware to force logging of errors even when Debug = True.  You'll
find this useful in the case that you have QA in DEBUG mode, and you'd still
like to log exceptions that show up there (not just to the screen, which is the
default behavior in DEBUG mode)

If you don't want to fiddle with your middlewares list in different
environments, you can just add this permanently, and then easily turn it off by
using this setting: 
DEBUG_MODE_LOGGING = False

You might find it useful to do that for local development, since it may get
annoying to wade through exception logs on your console when you're already
seeing every error on the screen.
    """

    def __init__(self):
        super(DebugModeLoggingMiddleware,self).__init__()
        self.log = logger.get_log(__name__)
        if not getattr(settings, 'DEBUG', False):
            self.log.info('DebugModeLoggingMiddleware has been turned off for all requests cuz we\'re not in debug mode')
            raise MiddlewareNotUsed
        if not getattr(settings, 'DEBUG_MODE_LOGGING', True):
            self.log.info('DebugModeLoggingMiddleware has been explicitly turned off for all requests')
            raise MiddlewareNotUsed
        self.log.info('DebugModeLoggingMiddleware has been activated')

    def process_exception(self, request, exception):
        if settings.DEBUG:
            self.log.error(traceback.format_exc(exception))




"""
The RequestSupplement object that will get injected on marketplace requests
"""

class RequestSupplement(object):
    # adding some duplicate fields keyed to more consistent python naming and
    # more understandable naming in some cases.  You're free to use whichever
    # you want.

    EXTRAS = { 
        'portal_id': 'hub_id',
        'app_pageUrl': 'local_url',
        'app_callbackUrl': 'local_base_url',
        'app_canvasUrl': 'base_url',
    }

    def __init__(self, request):
        super(RequestSupplement,self).__init__()
        self.process(request)

    def process(self, request):
        for k in request.REQUEST:
            if k.startswith('hubspot.marketplace.'):
                attr = '_'.join(k.split('.')[2:])
                val = request.REQUEST.get(k)
                if attr.endswith('_id'):
                    val = long(val)
                elif attr.startswith('is_'):
                    val = val.lower()=='true'
                setattr(self, attr, val)
        for k in self.__class__.EXTRAS:
            if getattr(self,k,None): # they're not all necessarily here (uninstall hook only gives secret and portal_id)
                setattr(self,self.__class__.EXTRAS[k],getattr(self,k))





#!/usr/bin/env python
from setuptools import setup, find_packages

VERSION = '1.6.1'

setup(
    name='hspy',
    version=VERSION,
    author='prior',
    author_email='mprior@hubspot.com',
    packages=find_packages(),
    url='https://github.com/HubSpot/hspy',
    download_url='https://github.com/HubSpot/hspy/tarball/v%s'%VERSION,
    license='LICENSE.txt',
    description='Utilities to aide HubSpot Marketplace development',
    long_description=open('README.rst').read(),
    install_requires=[
        'Django==1.3'
    ],
    platforms=['any']
)

